#include "mbed.h"
#include "LSM9DS0.h"
#include <math.h>
#include <cstdlib>
#include <cmath>
#include <limits>

DigitalOut myled(D13);

// timer 1
#define LOOPTIME  10000                         // 10 ms
unsigned long lastMilli = 0;

Timer t;

Serial uart_1(D10,D2);           // TX : D10 , RX : D2         // serial 1

void init_uart(void);
void separate(void);
void uart_send(void);

float lpf(float input, float output_old, float frequency);
double generateGaussianNoise(double mu, double sigma);

// uart send data
int display[6] = {0,0,0,0,0,0};
char num[14] = {254,255,0,0,0,0,0,0,0,0,0,0,0,0};      // char num[0] , num[1] : 2 start byte

void init_sensor(void);
void read_sensor(void);

LSM9DS0 sensor(SPI_MODE, D7, D6);    // SPI_CS1 : D7 , SPI_CS2 : D6

int sensor_flag = 0;                 // sensor initial flag
int sensor_count = 0;

// sensor data
int16_t Gyro_axis_data[3] = {0};     // X, Y, Z axis
int16_t Acce_axis_data[3] = {0};     // X, Y, Z axis
int16_t Magn_axis_data[3] = {0};     // X, Y, Z axis

// sampling time
float T = 0.010;                     // 10 ms

#define r_01 ((double)rand()/(RAND_MAX))            // uniform distribution 0 ~ 1
#define r_lu ((double)rand()/(RAND_MAX)*2-1)        // uniform distribution -1 ~ 1

// particle filter variables
float x = 0.1;          // initial actual state
float x_N = 0.1;          // process noise variance
float x_R = 0.1;        // measurement noise variance
#define N 50           // particle number
float V = 1;            // initial variance
float x_P[N];           // initial perticles
float x_P_updata[N];
float y_updata[N];
float w_P[N];
float sum_w_P = 0;
float w_P_n[N];
float w_P_n_c[N];
float y = 0;
float sum_x_P = 0;
float x_est = 0;

float Gyro_axis_data_f[3];
float Gyro_axis_data_f_old[3];

int main()
{
    t.start();

    init_uart();
    init_sensor();
    
    myled = 1;
    
    // generate initial particles
    for(int z=0 ; z<N ; z++)
    {
        x_P[z] = x + sqrt(V)*generateGaussianNoise(0,1);
    }

    while(1) 
    {
        // timer 1
        if((t.read_us() - lastMilli) >= LOOPTIME)  
        {    
            lastMilli = t.read_us();
            
            myled = !myled;

            // sensor initial start
            if(sensor_flag == 0) 
            {
                sensor_count++;

                if(sensor_count >= 200) 
                {
                    sensor_flag  = 1;
                    sensor_count = 0;
                }
            } 
            else 
            {
                read_sensor();
                uart_send();
            }
            
            // particle filter start
            // model : x = x + v , measurement : y = x + n , assume v n are Gaussian distribution
            // measurement
            y = Gyro_axis_data[0]*0.00122173;
            // lpf
            Gyro_axis_data_f[0] = lpf(y,Gyro_axis_data_f_old[0],10);     
            Gyro_axis_data_f_old[0] = Gyro_axis_data_f[0];       
            // sampling and calculate weights
            for(int e=0 ; e<N ; e++)
            {
                x_P_updata[e] = x_P[e] + sqrt(x_N)*generateGaussianNoise(0,1);
                y_updata[e] = x_P_updata[e];
                w_P[e] = (1/sqrt(2*PI*x_R)) * exp(-pow((y - y_updata[e]),2)/(2*x_R)); 
            }
            // sum of all weights
            float sum_w_P_ = 0;
            for(int f=0 ; f<N ; f++)
            {
                sum_w_P_ = sum_w_P_ + w_P[f];
            }
            sum_w_P = sum_w_P_;
            // normalize all weights
            for(int g=0 ; g<N ; g++)
            {
                w_P_n[g] = w_P[g] / sum_w_P;
            }
            // accumulate all weights
            for(int h=0 ; h<N ; h++)
            {
                if(h >= 1)
                {
                    w_P_n_c[h] = w_P_n_c[h-1] + w_P_n[h];
                }
                else
                {
                    w_P_n_c[h] = w_P_n[h];
                }
            }
            // resampling
            for(int a=0 ; a<N ; a++)
            {
                for(int b=0 ; b<N ; b++)
                {
                    if(w_P_n_c[b] > r_01)
                    {
                        x_P[a] = x_P_updata[b];
                        break;
                    }
                }
            }
            // average of all particles
            float sum_x_P_ = 0;
            for(int c=0 ; c<N ; c++)
            {
                sum_x_P_ = sum_x_P_ + x_P[c];
            }
            sum_x_P = sum_x_P_;
            
            x_est = sum_x_P / N;
        }    // timer 1 end
    }        // while(1) end
}

int i = 0;
void uart_send(void)
{
    // uart send data
    display[0] = y;
    display[1] = x_est;
    display[2] = Gyro_axis_data_f[0];
    display[3] = w_P_n_c[N-1];
    display[4] = 10*r_01;
    display[5] = 10*generateGaussianNoise(0,1);

    separate();

    int j = 0;
    int k = 1;
    while(k) 
    {
        if(uart_1.writeable()) 
        {
            uart_1.putc(num[i]);
            i = i + 1;
            j = j + 1;
        }

        if(j>1) 
        {
            k = 0;
            j = 0;
        }
    }

    if(i>13)
        i = 0;
}

void read_sensor(void)
{
    // sensor raw data
    Gyro_axis_data[0] = sensor.readRawGyroX();
    Gyro_axis_data[1] = sensor.readRawGyroY();
    Gyro_axis_data[2] = sensor.readRawGyroZ();

    Acce_axis_data[0] = sensor.readRawAccelX();
    Acce_axis_data[1] = sensor.readRawAccelY();
    Acce_axis_data[2] = sensor.readRawAccelZ();

    Magn_axis_data[0] = sensor.readRawMagX();
    Magn_axis_data[1] = sensor.readRawMagY();
    Magn_axis_data[2] = sensor.readRawMagZ();
}

void init_uart()
{
    uart_1.baud(115200);      // 設定baudrate
}

void separate(void)
{
    num[2] = display[0] >> 8;           // HSB(8bit)資料存入陣列
    num[3] = display[0] & 0x00ff;       // LSB(8bit)資料存入陣列
    num[4] = display[1] >> 8;
    num[5] = display[1] & 0x00ff;
    num[6] = display[2] >> 8;
    num[7] = display[2] & 0x00ff;
    num[8] = display[3] >> 8;
    num[9] = display[3] & 0x00ff;
    num[10] = display[4] >> 8;
    num[11] = display[4] & 0x00ff;
    num[12] = display[5] >> 8;
    num[13] = display[5] & 0x00ff;
}

void init_sensor(void)
{
    sensor.begin();
    // sensor.begin() setting :
    // uint16_t begin(gyro_scale gScl = G_SCALE_2000DPS,
    //                accel_scale aScl = A_SCALE_8G,
    //                mag_scale mScl = M_SCALE_8GS,
    //                gyro_odr gODR = G_ODR_760_BW_100,
    //                accel_odr aODR = A_ODR_800,
    //                mag_odr mODR = M_ODR_100);
}

float lpf(float input, float output_old, float frequency)
{
    float output = 0;
    
    output = (output_old + frequency*T*input) / (1 + frequency*T);
    
    return output;
}

double generateGaussianNoise(double mu, double sigma)
{
    const double epsilon = std::numeric_limits<double>::min();
    const double two_pi = 2.0*3.14159265358979323846;

    static double z0, z1;
    static bool generate;
    
    generate = !generate;

    if(!generate)
       return z1 * sigma + mu;

    double u1, u2;
    
    do
    {
       u1 = rand() * (1.0 / RAND_MAX);
       u2 = rand() * (1.0 / RAND_MAX);
    }
     
    while(u1 <= epsilon);

    z0 = sqrt(-2.0 * log(u1)) * cos(two_pi * u2);
    z1 = sqrt(-2.0 * log(u1)) * sin(two_pi * u2);
    
    return z0 * sigma + mu;
}
